# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/20_guitar_chords.ipynb.

# %% auto 0
__all__ = ['GuitarNote', 'GuitarSoundGenerator']

# %% ../../../nbs/20_guitar_chords.ipynb 2
from enum import Enum
from pathlib import Path
from dataclasses import dataclass
from typing import Literal, TypeAlias

import numpy as np
import soundfile as sf

# %% ../../../nbs/20_guitar_chords.ipynb 5
GuitarNote: TypeAlias = Literal['C', 'D', 'E', 'F', 'G', 'A', 'B']

# %% ../../../nbs/20_guitar_chords.ipynb 6
class GuitarSoundGenerator:
    """Simple guitar sound waveform generator."""

    # Note frequencies in Hz for octave 0
    NOTES = {
        'C': 32.7,
        'D': 36.71,
        'E': 41.2,
        'F': 43.65,
        'G': 49.0,
        'A': 27.5,
        'B': 30.87,
    }

    def __init__(self, sample_rate: int = 44100):
        """
        Initialize the note generator.

        Args:
            sample_rate: Determines how many discrete samples per second are 
                generated for the waveform.
        """
        self.sample_rate = sample_rate

    def get_frequency(self, note: GuitarNote, *, octave: int = 4) -> float:
        """
        Get frequency for a note in a specific octave.
        
        Args:
            note: Note name (e.g., 'A', 'G')
            octave: Determines the pitch register. Each octave increase doubles
                the frequency. For example A0 = 27.5Hz, A1 = 55Hz, A2 = 110Hz, 
                A3 = 220Hz and A4 = 440Hz

        Returns:
            Frequency in Hz
        """
        # A0 is in octave 0, so A4 = A0 * 2^4
        base_freq = self.NOTES[note]
        frequency = base_freq * (2 ** octave)
        return frequency

    def generate_note(
        self, 
        note: GuitarNote, 
        *,
        octave: int = 4, 
        duration: float = 1.0, 
        amplitude: float = 0.5,
    ) -> np.ndarray:
        """
        Generate a waveform for a musical note.
        
        Args:
            note: Note name (e.g., 'A', 'G')
            octave: Octave number
            duration: Duration in seconds
            amplitude: Amplitude (0.0 to 1.0)
            wave_type: Type of wave
        Returns:
            Numpy array containing the waveform
        """
        frequency = self.get_frequency(note, octave=octave)
        
        # numpy array that represent time points in seconds
        tp_in_s = np.linspace(
            0, 
            duration, 
            int(self.sample_rate * duration), 
            False
        )
        
        
        waveform = amplitude * np.sin(2 * np.pi * frequency * tp_in_s)

        # Apply fade in/out to avoid clicks
        fade_samples = int(0.01 * self.sample_rate)  # 10ms fade
        if len(waveform) > 2 * fade_samples:
            waveform[:fade_samples] *= np.linspace(0, 1, fade_samples)
            waveform[-fade_samples:] *= np.linspace(1, 0, fade_samples)
        
        return waveform

    def generate_chord(
        self, 
        notes: list[GuitarNote],
        *,
        octave: int = 4, 
        duration: float = 1.0,
        amplitude: float = 0.3,
    ) -> np.ndarray:
        """
        Generate a chord by combining multiple notes.

        Args:
            notes: List of note names
            octave: Octave number
            duration: Duration in seconds
            amplitude: Amplitude per note

        Returns:
            Numpy array containing the combined waveform
        """
        waveforms = []
        for note in notes:
            waveform = self.generate_note(
                note, 
                octave=octave, 
                duration=duration, 
                amplitude=amplitude
            )
            waveforms.append(waveform)

        # Sum all waveforms
        combined = np.sum(waveforms, axis=0)

        # Normalize to prevent clipping
        max_val = np.max(np.abs(combined))
        if max_val > 1.0:
            combined = combined / max_val

        return combined

    def to_wav(self, waveform: np.ndarray, file: str | Path) -> None:
        """
        Save waveform as WAV file.

        Args:
            waveform: Numpy array containing the waveform
            file_path: Output filename. Can be absolute or relative path.
                Automatically adds .wav extension if not present.
        """
        path = Path(file)

        if path.suffix.lower() != '.wav':
            path = path.with_suffix('.wav')

        path.parent.mkdir(parents=True, exist_ok=True)

        # Ensure waveform is in the correct range and format
        waveform_clipped = np.clip(waveform, -1.0, 1.0)

        # Save as 32-bit float WAV
        sf.write(str(path), waveform_clipped, self.sample_rate, subtype='FLOAT')



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/bioacoustics',
                'doc_host': 'https://itepifanio.github.io',
                'git_url': 'https://github.com/itepifanio/bioacoustics',
                'lib_path': 'bioacoustics_augmentation'},
  'syms': { 'bioacoustics_augmentation.augmentations.classical.audio': { 'bioacoustics_augmentation.augmentations.classical.audio.pitch_shift': ( 'classical_audio_augmentations.html#pitch_shift',
                                                                                                                                                  'bioacoustics_augmentation/augmentations/classical/audio.py'),
                                                                         'bioacoustics_augmentation.augmentations.classical.audio.time_shift': ( 'classical_audio_augmentations.html#time_shift',
                                                                                                                                                 'bioacoustics_augmentation/augmentations/classical/audio.py')},
            'bioacoustics_augmentation.augmentations.classical.spectrogram': { 'bioacoustics_augmentation.augmentations.classical.spectrogram._apply_time_warp': ( 'classical_spectrogram_augmentions.html#_apply_time_warp',
                                                                                                                                                                   'bioacoustics_augmentation/augmentations/classical/spectrogram.py'),
                                                                               'bioacoustics_augmentation.augmentations.classical.spectrogram._frequency_masking': ( 'classical_spectrogram_augmentions.html#_frequency_masking',
                                                                                                                                                                     'bioacoustics_augmentation/augmentations/classical/spectrogram.py'),
                                                                               'bioacoustics_augmentation.augmentations.classical.spectrogram._sparse_image_warp': ( 'classical_spectrogram_augmentions.html#_sparse_image_warp',
                                                                                                                                                                     'bioacoustics_augmentation/augmentations/classical/spectrogram.py'),
                                                                               'bioacoustics_augmentation.augmentations.classical.spectrogram._time_masking': ( 'classical_spectrogram_augmentions.html#_time_masking',
                                                                                                                                                                'bioacoustics_augmentation/augmentations/classical/spectrogram.py'),
                                                                               'bioacoustics_augmentation.augmentations.classical.spectrogram.apply_spec_augment': ( 'classical_spectrogram_augmentions.html#apply_spec_augment',
                                                                                                                                                                     'bioacoustics_augmentation/augmentations/classical/spectrogram.py'),
                                                                               'bioacoustics_augmentation.augmentations.classical.spectrogram.pitch_shift': ( 'classical_spectrogram_augmentions.html#pitch_shift',
                                                                                                                                                              'bioacoustics_augmentation/augmentations/classical/spectrogram.py'),
                                                                               'bioacoustics_augmentation.augmentations.classical.spectrogram.random_shifts': ( 'classical_spectrogram_augmentions.html#random_shifts',
                                                                                                                                                                'bioacoustics_augmentation/augmentations/classical/spectrogram.py'),
                                                                               'bioacoustics_augmentation.augmentations.classical.spectrogram.time_shift': ( 'classical_spectrogram_augmentions.html#time_shift',
                                                                                                                                                             'bioacoustics_augmentation/augmentations/classical/spectrogram.py')},
            'bioacoustics_augmentation.augmentations.mix': { 'bioacoustics_augmentation.augmentations.mix._paste_patch': ( 'combination_augmentations.html#_paste_patch',
                                                                                                                           'bioacoustics_augmentation/augmentations/mix.py'),
                                                             'bioacoustics_augmentation.augmentations.mix._rand_bbox': ( 'combination_augmentations.html#_rand_bbox',
                                                                                                                         'bioacoustics_augmentation/augmentations/mix.py'),
                                                             'bioacoustics_augmentation.augmentations.mix.cutmix': ( 'combination_augmentations.html#cutmix',
                                                                                                                     'bioacoustics_augmentation/augmentations/mix.py'),
                                                             'bioacoustics_augmentation.augmentations.mix.mixup': ( 'combination_augmentations.html#mixup',
                                                                                                                    'bioacoustics_augmentation/augmentations/mix.py')},
            'bioacoustics_augmentation.core': { 'bioacoustics_augmentation.core.foo': ( 'core.html#foo',
                                                                                        'bioacoustics_augmentation/core.py')},
            'bioacoustics_augmentation.utils.audio': { 'bioacoustics_augmentation.utils.audio.AudioMetadata': ( 'audio.html#audiometadata',
                                                                                                                'bioacoustics_augmentation/utils/audio.py'),
                                                       'bioacoustics_augmentation.utils.audio.Spectrogram': ( 'audio.html#spectrogram',
                                                                                                              'bioacoustics_augmentation/utils/audio.py'),
                                                       'bioacoustics_augmentation.utils.audio.Spectrogram.__init__': ( 'audio.html#spectrogram.__init__',
                                                                                                                       'bioacoustics_augmentation/utils/audio.py'),
                                                       'bioacoustics_augmentation.utils.audio.Spectrogram.copy': ( 'audio.html#spectrogram.copy',
                                                                                                                   'bioacoustics_augmentation/utils/audio.py'),
                                                       'bioacoustics_augmentation.utils.audio.Spectrogram.freq_bins': ( 'audio.html#spectrogram.freq_bins',
                                                                                                                        'bioacoustics_augmentation/utils/audio.py'),
                                                       'bioacoustics_augmentation.utils.audio.Spectrogram.time_frames': ( 'audio.html#spectrogram.time_frames',
                                                                                                                          'bioacoustics_augmentation/utils/audio.py'),
                                                       'bioacoustics_augmentation.utils.audio.Wav': ( 'audio.html#wav',
                                                                                                      'bioacoustics_augmentation/utils/audio.py'),
                                                       'bioacoustics_augmentation.utils.audio.Wav.__init__': ( 'audio.html#wav.__init__',
                                                                                                               'bioacoustics_augmentation/utils/audio.py'),
                                                       'bioacoustics_augmentation.utils.audio.Wav.copy': ( 'audio.html#wav.copy',
                                                                                                           'bioacoustics_augmentation/utils/audio.py'),
                                                       'bioacoustics_augmentation.utils.audio.Wav.from_spectrogram': ( 'audio.html#wav.from_spectrogram',
                                                                                                                       'bioacoustics_augmentation/utils/audio.py'),
                                                       'bioacoustics_augmentation.utils.audio.Wav.load': ( 'audio.html#wav.load',
                                                                                                           'bioacoustics_augmentation/utils/audio.py'),
                                                       'bioacoustics_augmentation.utils.audio.Wav.to_spectrogram': ( 'audio.html#wav.to_spectrogram',
                                                                                                                     'bioacoustics_augmentation/utils/audio.py'),
                                                       'bioacoustics_augmentation.utils.audio.play_audio': ( 'audio.html#play_audio',
                                                                                                             'bioacoustics_augmentation/utils/audio.py')},
            'bioacoustics_augmentation.utils.generator.guitar_chords': { 'bioacoustics_augmentation.utils.generator.guitar_chords.GuitarSoundGenerator': ( 'guitar_chords.html#guitarsoundgenerator',
                                                                                                                                                           'bioacoustics_augmentation/utils/generator/guitar_chords.py'),
                                                                         'bioacoustics_augmentation.utils.generator.guitar_chords.GuitarSoundGenerator.__init__': ( 'guitar_chords.html#guitarsoundgenerator.__init__',
                                                                                                                                                                    'bioacoustics_augmentation/utils/generator/guitar_chords.py'),
                                                                         'bioacoustics_augmentation.utils.generator.guitar_chords.GuitarSoundGenerator.generate_chord': ( 'guitar_chords.html#guitarsoundgenerator.generate_chord',
                                                                                                                                                                          'bioacoustics_augmentation/utils/generator/guitar_chords.py'),
                                                                         'bioacoustics_augmentation.utils.generator.guitar_chords.GuitarSoundGenerator.generate_note': ( 'guitar_chords.html#guitarsoundgenerator.generate_note',
                                                                                                                                                                         'bioacoustics_augmentation/utils/generator/guitar_chords.py'),
                                                                         'bioacoustics_augmentation.utils.generator.guitar_chords.GuitarSoundGenerator.get_frequency': ( 'guitar_chords.html#guitarsoundgenerator.get_frequency',
                                                                                                                                                                         'bioacoustics_augmentation/utils/generator/guitar_chords.py'),
                                                                         'bioacoustics_augmentation.utils.generator.guitar_chords.GuitarSoundGenerator.to_wav': ( 'guitar_chords.html#guitarsoundgenerator.to_wav',
                                                                                                                                                                  'bioacoustics_augmentation/utils/generator/guitar_chords.py')},
            'bioacoustics_augmentation.utils.image': { 'bioacoustics_augmentation.utils.image.BoundingBox': ( 'image.html#boundingbox',
                                                                                                              'bioacoustics_augmentation/utils/image.py'),
                                                       'bioacoustics_augmentation.utils.image.BoundingBox.area': ( 'image.html#boundingbox.area',
                                                                                                                   'bioacoustics_augmentation/utils/image.py')},
            'bioacoustics_augmentation.utils.plot': { 'bioacoustics_augmentation.utils.plot.plot_spectrogram': ( 'plot.html#plot_spectrogram',
                                                                                                                 'bioacoustics_augmentation/utils/plot.py'),
                                                      'bioacoustics_augmentation.utils.plot.plot_waveform': ( 'plot.html#plot_waveform',
                                                                                                              'bioacoustics_augmentation/utils/plot.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/12_image.ipynb.

# %% auto 0
__all__ = ['BoundingBox']

# %% ../../nbs/12_image.ipynb 2
from dataclasses import dataclass

# %% ../../nbs/12_image.ipynb 4
@dataclass(frozen=True)
class BoundingBox:
    """
    Bounding box representation with:
        - (x1, y1) been the top-left corner
        - (x2, y2) been the bottom-right corner
    """
    x1: int
    y1: int
    x2: int
    y2: int

    def area(self) -> int:
        width = self.x2 - self.x1
        height = self.y2 - self.y1
        return width * height



"""This notebooks presents classical bioacoustic audio augmenting techniques"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/01_classical_audio_augmentations.ipynb.

# %% auto 0
__all__ = ['pitch_shift', 'time_shift']

# %% ../../../nbs/01_classical_audio_augmentations.ipynb 2
import librosa
import numpy as np
from typing import Literal, Any, assert_never
import python_stretch

# %% ../../../nbs/01_classical_audio_augmentations.ipynb 6
def pitch_shift(
    wav_data: np.ndarray,
    sample_rate: int,
    n_steps: float,
    method: Literal['librosa', 'python_stretch'] = 'python_stretch'
) -> np.ndarray:
    """
    Applies pitch shift to the wav data.

    Args:
        wav_data: Audio data from a wav file.
        sample_rate: Sample rate of the audio.
        n_steps: Number of semitones to shift (positive = higher, negative = lower).
        method: librosa or pitch strech

    Returns:
        np.ndarray: Pitch-shifted wav data.
    """
    if method == 'librosa':
        return librosa.effects.pitch_shift(wav_data, sr=sample_rate, n_steps=n_steps)
    elif method == 'python_stretch':
        ndim = wav_data.ndim
        data = wav_data.copy()
        if ndim == 1:
            data = data[np.newaxis, :]

        stretch = python_stretch.Signalsmith.Stretch()
        stretch.preset(data.shape[0], sample_rate)
        stretch.setTransposeSemitones(n_steps)
        samples = stretch.process(data)

        return samples[0] if samples.ndim > ndim else samples

    assert_never()

# %% ../../../nbs/01_classical_audio_augmentations.ipynb 17
def time_shift(
    wav_data: np.ndarray,
    sample_rate: int,
    rate: float,
    method: Literal['librosa', 'python_stretch'] = 'python_stretch'
) -> np.ndarray:
    """
    Applies pitch shift to the wav data.

    Args:
        wav_data: Audio data from a wav file.
        sample_rate: Sample rate of the audio.
        rate: rate to be shifted.
        method: librosa or pitch strech

    Returns:
        np.ndarray: Pitch-shifted wav data.
    """
    if method == 'librosa':
        return librosa.effects.time_stretch(wav_data, rate=rate)
    elif method == 'python_stretch':
        ndim = wav_data.ndim
        data = wav_data.copy()
        if ndim == 1:
            data = data[np.newaxis, :]

        stretch = python_stretch.Signalsmith.Stretch()
        stretch.preset(data.shape[0], sample_rate)
        stretch.setTimeFactor(rate)
        samples = stretch.process(data)

        return samples[0] if samples.ndim > ndim else samples

    assert_never()



"""This notebook presents augmentation techniques that combine multiple more than one sample: Mixup, CutMix, SpecMix and more"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/03_combination_augmentations.ipynb.

# %% auto 0
__all__ = ['mixup', 'cutmix']

# %% ../../nbs/03_combination_augmentations.ipynb 2
import copy
import warnings
from typing import TypeAlias, Sequence

import numpy as np
from numpy.typing import NDArray

from bioacoustics_augmentation.utils.audio import Spectrogram, AudioMetadata
from bioacoustics_augmentation.utils.image import BoundingBox

# %% ../../nbs/03_combination_augmentations.ipynb 5
FloatArray: TypeAlias = NDArray[np.floating]
IntArray: TypeAlias = NDArray[np.integer]

# %% ../../nbs/03_combination_augmentations.ipynb 6
def mixup(
    primary_spec: FloatArray,
    secondary_spec: FloatArray,
    primary_label: FloatArray,
    secondary_label: FloatArray,
    alpha: float = 1.0,
    *,
    random_generator: np.random.Generator | None = None,
) -> tuple[FloatArray, FloatArray, float]:
    if random_generator is None:
        random_generator = np.random.default_rng()

    if not (alpha and alpha > 0):
        lambda_ = 1.0
        warnings.warn("Mixup is being used with invalid alpha. No Mixup will be applied.")
    else:
        lambda_ = random_generator.beta(alpha, alpha)

    mixed_data = lambda_ * primary_spec + (1.0 - lambda_) * secondary_spec
    mixed_label = lambda_ * primary_label + (1.0 - lambda_) * secondary_label

    return mixed_data, mixed_label, lambda_

# %% ../../nbs/03_combination_augmentations.ipynb 8
def _rand_bbox(
    size: tuple[int, ...],
    lambda_: float,
    random_generator: np.random.Generator
) -> BoundingBox:
    """
    Generate random bounding box coordinates for CutMix
    """
    width = size[-1]
    height = size[-2] if len(size) > 1 else 1

    cut_ratio = np.sqrt(1.0 - lambda_)
    cut_width = int(width * cut_ratio)
    cut_height = int(height * cut_ratio)

    # Uniformly sample the center of the patch
    center_x = random_generator.integers(0, width)
    center_y = random_generator.integers(0, height)

    x1 = np.clip(center_x - cut_width // 2, 0, width)
    y1 = np.clip(center_y - cut_height // 2, 0, height)
    x2 = np.clip(center_x + cut_width // 2, 0, width)
    y2 = np.clip(center_y + cut_height // 2, 0, height)

    return BoundingBox(x1=x1, y1=y1, x2=x2, y2=y2)

# %% ../../nbs/03_combination_augmentations.ipynb 9
def _paste_patch(
    source_sample: FloatArray,
    destination_sample: FloatArray,
    bbox: BoundingBox
) -> None:
    """
    Pastes a patch from the source sample to the destination sample.
    """
    if destination_sample.ndim == 1:
        # Handles 1D data like sequences (Batch, Features)
        destination_sample[bbox.x1:bbox.x2] = source_sample[bbox.x1:bbox.x2]
    else:
        # Handles 2D+ data like images (Batch, Channels, Height, Width)
        slicing = (..., slice(bbox.y1, bbox.y2), slice(bbox.x1, bbox.x2))
        destination_sample[slicing] = source_sample[slicing]

# %% ../../nbs/03_combination_augmentations.ipynb 10
def cutmix(
    primary_spec: FloatArray,
    secondary_spec: FloatArray,
    primary_label: FloatArray,
    secondary_label: FloatArray,
    alpha: float = 1.0,
    *,
    random_generator: np.random.Generator | None = None,
) -> tuple[FloatArray, FloatArray, float]:
    if random_generator is None:
        random_generator = np.random.default_rng()

    if not (alpha and alpha > 0):
        warnings.warn("CutMix is used with an invalid alpha. No augmentation will be applied.")
        return primary_spec, primary_label, 1.0

    lambda_ = random_generator.beta(alpha, alpha)
    mixed_data = primary_spec.copy()

    bbox = _rand_bbox(primary_spec.shape, lambda_, random_generator)

    if primary_spec.ndim == 1:
        total_area = primary_spec.shape[-1]
    else:
        total_area = primary_spec.shape[-1] * primary_spec.shape[-2]

    if total_area > 0:
        actual_lambda = 1.0 - (bbox.area() / total_area)

        _paste_patch(
            destination_sample=mixed_data,
            source_sample=secondary_spec,
            bbox=bbox
        )

        mixed_label = actual_lambda * primary_label + (1.0 - actual_lambda) * secondary_label
    else:
        mixed_label = primary_label.copy()

    return mixed_data, mixed_label, lambda_



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/11_audio.ipynb.

# %% auto 0
__all__ = ['AudioMetadata', 'Spectrogram', 'Wav', 'play_audio']

# %% ../../nbs/11_audio.ipynb 2
import copy
import tempfile
from pathlib import Path
from dataclasses import dataclass
from typing import Self, assert_never

import librosa
import soundfile
import numpy as np

from IPython.display import Audio, display

# %% ../../nbs/11_audio.ipynb 5
@dataclass
class AudioMetadata:
    """
    Metadata for structured audio data.

    It uses librosa default values.
    """
    source_file: str | Path | None = None
    sample_rate: int = 22050
    duration: float | None = None
    n_fft: int = 2048
    hop_length: int = 512

# %% ../../nbs/11_audio.ipynb 6
class Spectrogram:
    """Class to handle spectrogram data in 2D format."""

    def __init__(
        self,
        data: np.ndarray,
        stft: np.ndarray | None = None,
        metadata: AudioMetadata | None = None
    ) -> None:
        if data.ndim != 2:
            raise ValueError(f"Spectrogram must be 2D array, got {data.ndim}D")

        self.data = data
        self.stft = stft
        self.metadata = metadata or AudioMetadata()

    @property
    def freq_bins(self) -> int:
        """Get number of frequency bins."""
        return self.data.shape[0]

    @property
    def time_frames(self) -> int:
        """Get number of time frames."""
        return self.data.shape[1]

    def copy(self) -> 'Spectrogram':
        """
        Create a deep copy of the Spectrogram object.

        Returns:
            New Spectrogram object with copied data and metadata.
        """
        return Spectrogram(
            data=self.data.copy(),
            stft=self.stft.copy() if self.stft is not None else None,
            metadata=copy.deepcopy(self.metadata)
        )

# %% ../../nbs/11_audio.ipynb 7
class Wav:
    def __init__(
        self,
        *,
        waveform: np.ndarray,
        metadata: AudioMetadata,
    ) -> None:
        if waveform.ndim != 1:
            raise ValueError(f"Waveform must be 1D array, got {waveform.ndim}D")

        self.waveform = waveform
        self.metadata = metadata

        if self.metadata.duration is None:
            self.metadata.duration = len(waveform) / metadata.sample_rate

    def to_spectrogram(self) -> Spectrogram:
        # Short-time Fourier transform
        stft = librosa.stft(
            self.waveform,
            n_fft=self.metadata.n_fft,
            hop_length=self.metadata.hop_length
        )

        magnitude = librosa.amplitude_to_db(np.abs(stft), ref=np.max)

        return Spectrogram(
            data=magnitude,
            stft=stft,
            metadata=self.metadata
        )

    @classmethod
    def from_spectrogram(cls, spectrogram: Spectrogram) -> Self:
        if spectrogram.stft is None:
            raise ValueError("Cannot reconstruct audio: STFT data is missing")

        # Use phase information for accurate reconstruction
        reconstructed_waveform = librosa.istft(
            spectrogram.stft,
            hop_length=spectrogram.metadata.hop_length
        )

        new_metadata = AudioMetadata(
            source_file=f"{spectrogram.metadata.source_file}_reconstructed" if spectrogram.metadata.source_file else "reconstructed",
            sample_rate=spectrogram.metadata.sample_rate,
            duration=len(reconstructed_waveform) / spectrogram.metadata.sample_rate,
            n_fft=spectrogram.metadata.n_fft,
            hop_length=spectrogram.metadata.hop_length
        )

        return cls(waveform=reconstructed_waveform, metadata=new_metadata)

    @classmethod
    def load(
        cls,
        data: str | Path | np.ndarray,
        metadata: AudioMetadata | None = None,
    ) -> Self:
        if metadata is None:
            metadata = AudioMetadata()

        if isinstance(data, (str, Path)):
            waveform, actual_sr = librosa.load(data, sr=metadata.sample_rate)

            metadata.sample_rate = int(actual_sr)
            metadata.source_file = str(data)
        elif isinstance(data, np.ndarray):
            waveform = data
        else:
            assert_never()

        return cls(waveform=waveform, metadata=metadata)

    def copy(self) -> 'Wav':
        """
        Create a deep copy of the Wav object.

        Returns:
            New Wav object with copied waveform and metadata.
        """
        return Wav(
            waveform=self.waveform.copy(),
            metadata=copy.deepcopy(self.metadata)
        )

# %% ../../nbs/11_audio.ipynb 8
def play_audio(
    audio_data: np.ndarray | str | Path | Spectrogram | Wav,
    sample_rate: int | None = None
) -> None:
    """
    Universal function to play audio from various input types.
    """
    if isinstance(audio_data, (str, Path)):
        # File path - use directly
        display(Audio(str(audio_data)))

    elif isinstance(audio_data, Wav):
        # Wav object - use waveform and sample rate
        display(Audio(audio_data.waveform, rate=audio_data.metadata.sample_rate))

    elif isinstance(audio_data, Spectrogram):
        if audio_data.stft is not None:
            # Spectrogram object - convert to wav first
            wav_obj = Wav.from_spectrogram(audio_data)
            display(Audio(wav_obj.waveform, rate=wav_obj.metadata.sample_rate))
        else:
            # No STFT -> estimate phase with Griffin-Lim from magnitude (data is in dB)
            # This yields an approximate reconstruction suitable for playback.
            mag = librosa.db_to_amplitude(audio_data.data)  # convert dB -> linear magnitude
            n_fft = audio_data.metadata.n_fft
            hop_length = audio_data.metadata.hop_length
            
            # Use improved Griffin-Lim parameters for better reconstruction
            waveform = librosa.griffinlim(
                mag, 
                n_iter=128,
                hop_length=hop_length, 
                win_length=n_fft,
                length=None  # Let it determine the length from spectrogram
            )
            display(Audio(waveform, rate=audio_data.metadata.sample_rate))

    elif isinstance(audio_data, np.ndarray):
        # Numpy array - need sample rate
        if sample_rate is None:
            raise ValueError("sample_rate is required when playing numpy arrays")
        display(Audio(audio_data, rate=sample_rate))

    else:
        raise TypeError(f"Unsupported audio data type: {type(audio_data)}")



"""This notebooks presents classical bioacoustic augmenting techniques, i.e, methods that manipulates the sound or spectogram."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/02_classical_spectrogram_augmentions.ipynb.

# %% auto 0
__all__ = ['time_shift', 'pitch_shift', 'random_shifts', 'apply_spec_augment']

# %% ../../../nbs/02_classical_spectrogram_augmentions.ipynb 2
import scipy
import librosa
import numpy as np
from typing import Literal, Any, assert_never
from tps import ThinPlateSpline

# %% ../../../nbs/02_classical_spectrogram_augmentions.ipynb 5
def time_shift(spectrogram: Spectrogram, shift: int):
    shifted_spectrogram = spectrogram.copy()

    freq_bins, time_frames = shifted_spectrogram.data.shape
    spectrogram1 = shifted_spectrogram.data[:, :time_frames - shift]
    spectrogram2 = shifted_spectrogram.data[:, time_frames - shift:]

    shifted_spectrogram.data = np.concatenate((spectrogram2, spectrogram1), axis=1)

    return shifted_spectrogram

# %% ../../../nbs/02_classical_spectrogram_augmentions.ipynb 6
def pitch_shift(spectrogram: Spectrogram, shift: int):
    shifted_spectrogram = spectrogram.copy()

    freq_bins, time_frames = shifted_spectrogram.data.shape
    spectrogram1 = shifted_spectrogram.data[:freq_bins - shift, :]
    spectrogram2 = shifted_spectrogram.data[freq_bins - shift:, :]
    shifted_spectrogram.data = np.concatenate((spectrogram2, spectrogram1), axis=0)

    return shifted_spectrogram

# %% ../../../nbs/02_classical_spectrogram_augmentions.ipynb 17
def random_shifts(
    spectrogram: Spectrogram,
    time_shift_range: tuple[int, int] | None = None,
    pitch_shift_range: tuple[int, int] | None = None,
) -> Spectrogram:
    """
    Randomly applies shift (both in time and pitch) to the original spectrogram.

    The amount of the time and pitch shifts is randomly picked from the
    ranges time_shift_range and pitch_shift_range.

    Parameters
    ----------
    spectrograms: Spectrogram
        Input spectrograms organized as a list or single Spectrogram object.
    time_shift_range: tuple[int, int] | None
        Min and max values for time shift represented as [min, max].
    pitch_shift_range: tuple[int, int] | None
        Min and max values for pitch shift represented as [min, max].

    Returns
    -------
    new_spectrograms : Spectrogram
        Spectrogram set of shifted spectrograms.
    """
    if time_shift_range is None:
        time_shift_range = [0, 0]

    if pitch_shift_range is None:
        pitch_shift_range = [0, 0]

    min_pitch_shift, max_pitch_shift = pitch_shift_range
    rand_pitch_shift = min_pitch_shift + np.random.rand() * (max_pitch_shift - min_pitch_shift)
    rand_pitch_shift = int(rand_pitch_shift)

    min_time_shift, max_time_shift = time_shift_range
    rand_time_shift = np.random.randint(min_time_shift, max_time_shift + 1)

    shifted_spectrogram = time_shift(spectrogram, rand_time_shift)
    shifted_spectrogram = pitch_shift(shifted_spectrogram, rand_pitch_shift)

    return shifted_spectrogram

# %% ../../../nbs/02_classical_spectrogram_augmentions.ipynb 29
def _sparse_image_warp(
    spectrogram: Spectrogram,
    source_control_point_locations: np.ndarray,
    dest_control_point_locations: np.ndarray,
    regularization_weight: float = 0.0,
) -> Spectrogram:
    """
    Sparse image warping using correspondences between sparse control points.

    It applies a non-linear warp to the image, where the warp is specified by
    the source and destination locations of a number of control points.

    Args:
        spectrogram: The input Spectrogram object.
        source_control_point_locations: An array of shape `[num_points, 2]` with source (y, x) coordinates.
        dest_control_point_locations: An array of shape `[num_points, 2]` with destination (y, x) coordinates.
        regularization_weight: Weight on smoothness regularizer in interpolation

    Returns:
        A tuple containing:
        - warped_spectrogram: The warped Spectrogram object.
        - flow_field: The dense flow field as a NumPy array of shape `[height, width, 2]`.
    """
    image_data = spectrogram.data
    height, width = image_data.shape[:2]

    # Handle 2D spectrogram data by adding a temporary channel dimension for warping logic
    if image_data.ndim == 2:
        image = image_data[..., np.newaxis]
    else:
        image = image_data

    tps = ThinPlateSpline(alpha=regularization_weight)
    tps.fit(dest_control_point_locations, source_control_point_locations)

    output_indices = np.indices((height, width), dtype=np.float32).transpose(1, 2, 0)
    source_indices = tps.transform(output_indices.reshape(-1, 2)).reshape(height, width, 2)

    map_coords = source_indices.transpose(2, 0, 1)

    warped_channels = [
        scipy.ndimage.map_coordinates(image[..., c], map_coords, order=1, mode='reflect')
        for c in range(image.shape[2])
    ]
    warped_image = np.stack(warped_channels, axis=-1)

    if image_data.ndim == 2:
        warped_image = np.squeeze(warped_image, axis=-1)

    warped_spectrogram = spectrogram.copy()
    warped_spectrogram.data = warped_image
    warped_spectrogram.stft = None  # Invalidate STFT as the phase is no longer correct

    return warped_spectrogram

# %% ../../../nbs/02_classical_spectrogram_augmentions.ipynb 30
def _apply_time_warp(
    spectrogram: Spectrogram,
    warp_param: int,
    random_generator: np.random.Generator
) -> Spectrogram:
    """
    Applies time warping by creating control points as described in the SpecAugment paper.
    """
    height, width = spectrogram.data.shape

    # Source point is a random point on the center horizontal line
    source_y = height // 2
    source_x = random_generator.integers(warp_param, width - warp_param)

    # Destination point is the source point shifted horizontally by a random amount
    dest_y = source_y
    dest_x = source_x + random_generator.integers(-warp_param, warp_param)

    # The paper mentions fixing 6 boundary points:
    # 4 corners and the middles of the vertical edges.
    # These points do not move.
    source_points = np.array([
        [0, 0], [height - 1, 0],                    # Left corners
        [0, width - 1], [height - 1, width - 1],    # Right corners
        [height // 2, 0], [height // 2, width - 1], # Vertical middles
        [source_y, source_x]                        # The point to be warped
    ])

    dest_points = np.array([
        [0, 0], [height - 1, 0],
        [0, width - 1], [height - 1, width - 1],
        [height // 2, 0], [height // 2, width - 1],
        [dest_y, dest_x]                            # The new location of the warped point
    ])

    return _sparse_image_warp(spectrogram, source_points, dest_points)

# %% ../../../nbs/02_classical_spectrogram_augmentions.ipynb 34
def _frequency_masking(
    spectrogram: Spectrogram,
    num_masks: int,
    mask_param: int,  # Frequency mask parameter (`F` as mentioned in the paper)
    mask_value: float,
    random_generator: np.random.Generator,
) -> Spectrogram:
    augmented_spec = spectrogram.copy()
    freq_bins, time_frames = augmented_spec.data.shape

    for _ in range(num_masks):
        # Choose a random mask width `f` from a uniform distribution [0, F)
        f = random_generator.integers(0, mask_param + 1)

        # Choose a random starting frequency bin `f0`
        if freq_bins - f > 0:
            f0 = random_generator.integers(0, freq_bins - f)

            # Apply the mask
            augmented_spec.data[f0:f0 + f, :] = mask_value

    return augmented_spec

# %% ../../../nbs/02_classical_spectrogram_augmentions.ipynb 38
def _time_masking(
    spectrogram: Spectrogram,
    num_masks: int,
    mask_param: int,  # Time mask parameter (`T` as mentioned in the paper)
    mask_value: float,
    random_generator: np.random.Generator,
) -> Spectrogram:
    augmented_spec = spectrogram.copy()
    _, time_frames = augmented_spec.data.shape

    for _ in range(num_masks):
        # Choose a random mask width `t` from a uniform distribution [0, T)
        t = random_generator.integers(0, mask_param + 1)

        # Choose a random starting time frame `t0`
        if time_frames - t > 0:
            t0 = random_generator.integers(0, time_frames - t)

            # Apply the mask
            augmented_spec.data[:, t0:t0 + t] = mask_value

    return augmented_spec

# %% ../../../nbs/02_classical_spectrogram_augmentions.ipynb 42
def apply_spec_augment(
    spectrogram: Spectrogram,
    time_warp_param: int = 80,
    freq_mask_param: int = 27,
    time_mask_param: int = 100,
    num_freq_masks: int = 1,
    num_time_masks: int = 1,
    mask_value: float = 0.0,
    random_generator: np.random.Generator | None = None,
) -> Spectrogram:
    """
    Applies SpecAugment transformations to a spectrogram using helper functions.

    Args:
        spectrogram: The input Spectrogram object.
        time_warp_param: The maximum time warp distance (W). Set to 0 to disable.
        freq_mask_param: The maximum width (F) of the frequency mask.
        time_mask_param: The maximum width (T) of the time mask.
        num_freq_masks: The number of frequency masks to apply.
        num_time_masks: The number of time masks to apply.
        mask_value: The value to fill the masked regions with. The paper uses the mean,
            which is 0 for normalized spectrograms.
        random_generator: A NumPy random generator for reproducibility.

    Returns:
        A new Spectrogram object with augmentations applied.
    """
    if random_generator is None:
        random_generator = np.random.default_rng()

    augmented_spec = spectrogram.copy()

    if time_warp_param > 0:
        augmented_spec = _apply_time_warp(
            spectrogram=augmented_spec,
            warp_param=time_warp_param,
            random_generator=random_generator,
        )

    augmented_spec = _frequency_masking(
        spectrogram=augmented_spec,
        num_masks=num_freq_masks,
        mask_param=freq_mask_param,
        mask_value=mask_value,
        random_generator=random_generator,
    )

    augmented_spec = _time_masking(
        spectrogram=augmented_spec,
        num_masks=num_time_masks,
        mask_param=time_mask_param,
        mask_value=mask_value,
        random_generator=random_generator,
    )

    return augmented_spec



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/10_plot.ipynb.

# %% auto 0
__all__ = ['plot_waveform', 'plot_spectrogram']

# %% ../../nbs/10_plot.ipynb 2
from pathlib import Path
from typing import assert_never

import librosa
import numpy as np
import matplotlib.pyplot as plt

from bioacoustics_augmentation.utils.audio import Spectrogram

# %% ../../nbs/10_plot.ipynb 4
def plot_waveform(
    waveform: np.ndarray,
    sample_rate: int,
) -> None:
    """
    Plots a waveform.

    Args:
        waveform: The audio waveform to plot.
        sample_rate: The sample rate of the audio waveform.
        title: The title of the plot.
    """
    librosa.display.waveshow(
        waveform,
        sr=sample_rate,
    )

# %% ../../nbs/10_plot.ipynb 5
def plot_spectrogram(
    spectrogram: np.ndarray | Spectrogram,
    *,
    sample_rate: int = 22050,
    fmax: int | None = None,
) -> None:
    """
    Plots a spectrogram.

    Args:
        spectrogram: The spectrogram to plot.
        sample_rate: Sample rate for proper frequency axis scaling.
        fmax: Maximum frequency to display (optional).
    """
    data: np.ndarray = spectrogram.data if isinstance(spectrogram, Spectrogram) else spectrogram

    if isinstance(spectrogram, np.ndarray):
        librosa.display.specshow(
            data,
            x_axis='time',
            y_axis='mel',
            fmax=fmax,
            sr=sample_rate,
        )
    elif isinstance(spectrogram, Spectrogram):
        librosa.display.specshow(
            spectrogram.data, 
            sr=spectrogram.metadata.sample_rate, 
            hop_length=spectrogram.metadata.hop_length, 
            x_axis='time', 
            y_axis='mel',
            fmax=fmax,
        )
    else:
        assert_never()
